/*                                */
/*  java2pad.c -- ᥤ롼  */
/*                                */
/*  java2pad  by S. Yoshida       */
/*                                */

#include "config.h"

#include <stdlib.h>
#include <string.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#include "java2pad.h"

token_type token = None;	/* ȡ */
char word[BUFSIZ] = {'\0'};	/* ɤʸ */

bool no_print_comment_flag = Off;	/* Ȥ򤽤ξɽʤ */
bool elseif_flag = Off;			/* else if ν椫 */

int tab = 4;			/*  */
bool optimize = Off;		/* -O ץ */
bool semicolon = Off;		/* -semicolon ץ */

bool debug = Off;		/* -debug ץ */
bool quiet = Off;		/* -quiet ץ */
bool print_warning = Off;	/* warningϤ뤫 */
int line_count = 1;		/* Ƥ */

FILE *input_stream;		/* ϥե */


#ifdef JAPANESE
static char kconv[BUFSIZ] = "nkf -e";	/* С */
static bool noconv = Off;		/* -noconv */
#endif
static bool nohelp = Off;		/* -nohelp */
static bool stdin_flag = Off;		/* -stdin */

static char filename[BUFSIZ] = {'\0'};		/* ϥե̾ */
static char command[BUFSIZ];			/* ޥ */

static char prepfile[BUFSIZ];		/* ץץåΥե̾ */

static void prep(void);			/* ץץåʸ \ ν */
static int p_getchar(FILE *);		/* ʸɤ߹ */

static void option(int,char **);	/* ץν */
static void print_help(void);		/* إץå */

#if defined(HAVE_ATEXIT)
static void at_exit(void);		/* λν */
#elif defined(HAVE_ON_EXIT)
static void at_exit(int,void *);	/* λν */
#endif

int main(int argc,char **argv)
{
    char tmp[BUFSIZ];
    char *env;

    /* exit() κݤν */
#if defined(HAVE_ATEXIT)
    atexit(at_exit);
#elif defined(HAVE_ON_EXIT)
    on_exit(at_exit,NULL);
#endif

    /* ץץåΥե̾ */
#ifdef TMP
    strcpy(prepfile,TMP);
    strcat(prepfile,"/");
#else
    strcpy(prepfile,"/tmp/");
#endif
    if((env = getenv("TMP")) != NULL  &&  strlen(env) > 0){
	/* Ķѿ TMP  */
	strcpy(prepfile,env);
	strcat(prepfile,"/");
    }
    strcat(prepfile,"java2pad-XXXXXX");

    option(argc,argv);			/* ץν */

    if(quiet == Off)
	fprintf(stderr,"\njava2pad %s  by %s\n\n",VERSION,AUTHOR);

    if(stdin_flag == On)
	filename[0] = '\0';

#ifdef JAPANESE
    if(noconv == Off)
	sprintf(command,"cat %s | %s",filename,kconv);
    else
#endif
	sprintf(command,"cat %s",filename);

    input_stream = popen(command,"r");

    mktemp(prepfile);

    prep();		/* ץץåʸ \ ν */

    pclose(input_stream);

    input_stream = fopen(prepfile,"r");

    parser();

    fclose(input_stream);

    remove(prepfile);

    if(quiet == Off)
	fprintf(stderr,"\n%d errors, %d warnings.\n",error_count,warning_count);

    return 0;
}

void prep(void)
/* ץץåʸ \ ν */
{
    FILE *fp;
    int ch;
    int pch = 0;

    fp = fopen(prepfile,"w");

    ch = p_getchar(input_stream);
    while(ch != EOF){
	if(ch == '\\'){
	    ch = p_getchar(input_stream);
	    if(ch != '\n'){
		fputc('\\',fp);
		fputc(ch,fp);
	    }
	}else
	    fputc(ch,fp);
	pch = ch;
	ch = p_getchar(input_stream);
    }

    fclose(fp);
}

int p_getchar(FILE *fp)
/* ʸɤ߹ */
{
    int c = 0;
    while(((0 <= c  &&  c < ' '  &&  c != '\t'  &&  c != '\n')  ||  c == 0x7F)  &&  c != EOF){
	c = getc(fp);
    }
    return c;
}

void option(int argc,char **argv)
/* ץν */
{
    int i,p;
    char *ptr;

    for(i = 1 ; i < argc ; i++){
	if(argv[i][0] != '-'){		/* ե̾ */
	    if(filename[0] == '\0')
		strcpy(filename,argv[i]);
	    else if(nohelp == Off){
		print_help();
		exit(1);
	    }
	}else{				/* ץ */
	    switch(argv[i][1]){
#ifdef DEBUG
		case 'd':
		    if(strcmp(argv[i],"-debug") == 0){
			debug = On;
		    }else if(nohelp == Off){
			print_help();
			exit(1);
		    }
		    break;
#endif
		case 'h':
		    if(strcmp(argv[i],"-help") == 0){
			print_help();
			exit(0);
		    }else if(nohelp == Off){
			print_help();
			exit(1);
		    }
		    break;
#ifdef JAPANESE
		case 'k':
		    if(strcmp(argv[i],"-kconv") == 0  &&  i + 1 < argc){
			strcpy(kconv,argv[++i]);
		    }else if(nohelp == Off){
			print_help();
			exit(1);
		    }
		    break;
#endif
		case 'n':
		    if(strcmp(argv[i],"-nohelp") == 0){
			nohelp = On;
		    }else if(strcmp(argv[i],"-noopt") == 0){
			optimize = Off;
#ifdef JAPANESE
		    }else if(strcmp(argv[i],"-noconv") == 0){
			noconv = On;
#endif
		    }else if(nohelp == Off){
			print_help();
			exit(1);
		    }
		    break;
		case 'O':
		    if(strcmp(argv[i],"-O") == 0){
			optimize = On;
		    }else if(nohelp == Off){
			print_help();
			exit(1);
		    }
		    break;
		case 'q':
		    if(strcmp(argv[i],"-quiet") == 0){
			quiet = On;
		    }else if(nohelp == Off){
			print_help();
			exit(1);
		    }
		    break;
		case 's':
		    if(strcmp(argv[i],"-semicolon") == 0){
			semicolon = On;
		    }else if(strcmp(argv[i],"-stdin") == 0){
			stdin_flag = On;
		    }else if(nohelp == Off){
			print_help();
			exit(1);
		    }
		    break;
		case 't':
		    if(strcmp(argv[i],"-tab") == 0){
			i++;
			tab = atoi(argv[i]);
		    }else if(nohelp == Off){
			print_help();
			exit(1);
		    }
		    break;
		case 'w':
		    if(strcmp(argv[i],"-warn") == 0){
			print_warning = On;
		    }else if(nohelp == Off){
			print_help();
			exit(1);
		    }
		    break;
		default:
		    if(nohelp == Off){
			print_help();
			exit(1);
		    }
		    break;
	    }
	}
    }
}

void print_help(void)
{
    fprintf(stderr,"\njava2pad %s  by %s\n\n",VERSION,AUTHOR);
    fputs("usage: java2pad [filename] [options]\n",stderr);
#ifdef DEBUG
    fputs("-debug                  Print parser debug message.\n",stderr);
#endif
    fputs("-help                   Print help message.\n",stderr);
#ifdef JAPANESE
    fputs("-kconv converter        Kanji code converter [default: \"nkf -e\"].\n",stderr);
    fputs("-noconv                 Do not convert kanji code.\n",stderr);
#endif
    fputs("-nohelp                 Do not print help message.\n",stderr);
    fputs("-noopt                  No optimizing.\n",stderr);
    fputs("-O                      Optimizing.\n",stderr);
    fputs("-quiet                  Do not print messages.\n",stderr);
    fputs("-semicolon              Print ';' at EOL.\n",stderr);
    fputs("-stdin                  Input from stdin.\n",stderr);
    fputs("-tab width              Set TAB width [default: 4].\n",stderr);
    fputs("-warn                   Print warnings.\n",stderr);
}


#if defined(HAVE_ATEXIT) || defined(HAVE_ON_EXIT)
#if defined(HAVE_ATEXIT)
void at_exit(void)
#elif defined(HAVE_ON_EXIT)
void at_exit(int val, void *p)
#endif
/* λν */
{
    remove(prepfile);
}
#endif
